@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -459,7 +459,7 @@ def manifest(url, renpy=False, insecure=False):
         with open(_path("temp:manifest.py.sig"), "rb") as f:
             sig = f.read()
 
-        key = ecdsa.VerifyingKey.from_pem(_renpy.exports.file("renpy_ecdsa_public.pem").read())
+        key = ecdsa.VerifyingKey.from_pem(_renpy.exports.open_file("renpy_ecdsa_public.pem").read())
 
         if not key.verify(sig, manifest):
             error(_("The manifest signature is not valid."))
