@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -134,7 +134,7 @@ style_properties = sorted_dict(
     alt=None,
     altruby_style=None,
     antialias=None,
-    vertical=None,
+    axis=None,
     background='renpy.easy.displayable_or_none',
     bar_invert=None,
     bar_resizing=None,
@@ -155,6 +155,8 @@ style_properties = sorted_dict(
     debug=None,
     drop_shadow=None,
     drop_shadow_color='renpy.easy.color',
+    emoji_font=None,
+    extra_alt=None,
     first_indent=None,
     first_spacing=None,
     fit_first=None,
@@ -164,10 +166,12 @@ style_properties = sorted_dict(
     fore_bar='none_is_null',
     fore_gutter=None,
     foreground='renpy.easy.displayable_or_none',
+    group_alt=None,
     hinting=None,
     hover_sound=None,
     hyperlink_functions=None,
     italic=None,
+    instance=None,
     justify=None,
     kerning=None,
     key_events=None,
@@ -187,10 +191,13 @@ style_properties = sorted_dict(
     order_reverse=None,
     outlines='expand_outlines',
     outline_scaling=None,
+    prefer_emoji=None,
     rest_indent=None,
     right_margin=None,
     right_padding=None,
+    ruby_line_leading=None,
     ruby_style=None,
+    shaper=None,
     size=None,
     size_group=None,
     slow_abortable=None,
@@ -209,6 +216,7 @@ style_properties = sorted_dict(
     top_margin=None,
     top_padding=None,
     underline=None,
+    vertical=None,
     xanchor='expand_anchor',
     xfill=None,
     xfit=None,
