@@ -1,4 +1,4 @@
-# Copyright 2004-2018 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -19,6 +19,11 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+from __future__ import division, absolute_import, with_statement, print_function, unicode_literals
+from renpy.compat import PY2, basestring, bchr, bord, chr, open, pystr, range, round, str, tobytes, unicode # *
+
+
+
 
 import renpy.display
 import colorsys
@@ -34,8 +39,8 @@ class Color(tuple):
     between various color spaces. It also represents opacity in the form
     of an alpha.
 
-    When creating a Color, at one of the `color`, `hls`, `hsv`, or `rgb`
-    arguments should be supplied. (If all are None, None is returned.)
+    When creating a Color, at most one of the `color`, `hls`, `hsv`, or
+    `rgb` arguments should be supplied. (If all are None, None is returned.)
 
     `color`
         The color, in one of the standard formats Ren'Py understands. These
@@ -94,6 +99,11 @@ class Color(tuple):
         the red, green, and blue components. Each component ranges between 0.0
         and 1.0.
 
+    .. attribute:: rgba
+
+        Returns the color as a tuple of four floating point numbers giving
+        the red, green, blue and alpha components as 0.0 to 1.0 values.
+
     .. attribute:: alpha
 
         Returns the alpha (opacity) of this Color as a number between 0.0 and
@@ -112,42 +122,35 @@ class Color(tuple):
     _hls = None
     _hsv = None
     _alpha = None
+    _rgba = None
 
     def __new__(cls, color=None, hls=None, hsv=None, rgb=None, alpha=1.0):
 
         if color is not None:
             c = color
 
-            if isinstance(c, tuple):
-                if isinstance(c, Color):
-                    return c
-
-                if len(c) == 4:
-                    return tuple.__new__(cls, c)
-
-                if len(c) == 3:
-                    return tuple.__new__(cls, c + (int(255 * alpha),))
-
             if isinstance(c, basestring):
                 if c[0] == '#':
                     c = c[1:]
 
-                if len(c) == 6:
-                    r = int(c[0]+c[1], 16)
-                    g = int(c[2]+c[3], 16)
-                    b = int(c[4]+c[5], 16)
+                lenc = len(c)
+
+                if lenc == 6:
+                    r = int(c[:2], 16)
+                    g = int(c[2:4], 16)
+                    b = int(c[4:], 16)
                     a = int(alpha * 255)
-                elif len(c) == 8:
-                    r = int(c[0]+c[1], 16)
-                    g = int(c[2]+c[3], 16)
-                    b = int(c[4]+c[5], 16)
-                    a = int(c[6]+c[7], 16)
-                elif len(c) == 3:
+                elif lenc == 8:
+                    r = int(c[:2], 16)
+                    g = int(c[2:4], 16)
+                    b = int(c[4:6], 16)
+                    a = int(c[6:], 16)
+                elif lenc == 3:
                     r = int(c[0], 16) * 0x11
                     g = int(c[1], 16) * 0x11
                     b = int(c[2], 16) * 0x11
                     a = int(alpha * 255)
-                elif len(c) == 4:
+                elif lenc == 4:
                     r = int(c[0], 16) * 0x11
                     g = int(c[1], 16) * 0x11
                     b = int(c[2], 16) * 0x11
@@ -155,7 +158,20 @@ class Color(tuple):
                 else:
                     raise Exception("Color string {!r} must be 3, 4, 6, or 8 hex digits long.".format(c))
 
-                return tuple.__new__(cls, (r, g, b, a))
+                return tuple.__new__(cls, (r, g, b, a)) # type: ignore
+
+            if isinstance(c, Color):
+                return c
+
+            c = tuple(c)
+
+            lenc = len(c)
+
+            if lenc == 4:
+                return tuple.__new__(cls, c)
+
+            if lenc == 3:
+                return tuple.__new__(cls, c + (int(255 * alpha),))
 
         if hsv is not None:
             rgb = colorsys.hsv_to_rgb(*hsv)
@@ -170,11 +186,12 @@ class Color(tuple):
             b = int(rgb[2] * 255)
             a = int(alpha * 255)
 
-            rv = tuple.__new__(cls, (r, g, b, a))
+            rv = tuple.__new__(cls, (r, g, b, a)) # type: ignore
             rv._rgb = rgb
             rv._hls = hls
             rv._hsv = hsv
             rv._alpha = alpha
+            rv._rgba = tuple(list(rgb) + [alpha])
 
             return rv
 
@@ -207,6 +224,18 @@ class Color(tuple):
 
         return self._rgb
 
+    @property
+    def rgba(self):
+        if self._rgba is None:
+            self._rgba = (
+                self[0] / 255.0,
+                self[1] / 255.0,
+                self[2] / 255.0,
+                self[3] / 255.0,
+                )
+
+        return self._rgba
+
     @property
     def hls(self):
         if self._hls is None:
@@ -280,7 +309,7 @@ class Color(tuple):
             self[2] * other[2],
             self[3] * other[3]))
 
-    __rmul__ = __mul__
+    __rmul__ = __mul__ # type: ignore
 
     def interpolate_core(self, a, b, fraction):
 
@@ -323,7 +352,7 @@ class Color(tuple):
         hsv = self.interpolate_core(self.hsv, other.hsv, fraction)
         alpha = self.interpolate_core(self.alpha, other.alpha, fraction)
 
-        return Color(hsv=hsv, alpha=alpha)
+        return Color(hsv=hsv, alpha=alpha) # type: ignore
 
     def interpolate_hls(self, other, fraction):
         """
@@ -344,7 +373,7 @@ class Color(tuple):
         hls = self.interpolate_core(self.hls, other.hls, fraction)
         alpha = self.interpolate_core(self.alpha, other.alpha, fraction)
 
-        return Color(hls=hls, alpha=alpha)
+        return Color(hls=hls, alpha=alpha) # type: ignore
 
     def tint(self, fraction):
         """
@@ -391,7 +420,8 @@ class Color(tuple):
         :doc: color method
 
         Rotates this color's hue by `rotation`, and returns the new Color. `rotation`
-        is a fraction of a full rotation, to convert degrees divide by 360.0.
+        is a fraction of a full rotation (between 0.0 and 1.0). Divide by 360.0 to
+        convert to degrees.
         """
 
         h, l, s = self.hls
