@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -391,6 +391,7 @@ adv = ADVCharacter(None,
                 callback=None,
                 type='say',
                 advance=True,
+                retain=False,
 
                 who_style='say_label',
                 what_style='say_dialogue',
@@ -401,8 +402,9 @@ adv = ADVCharacter(None,
 
                 kind=False)
 
-# This character is copied when a name-only say statement is called.
-name_only = adv
+# predict_say and who are defined in 00library.rpy, but we add default
+# versions here in case there is a problem with initialization. (And
+# for pickling purposes.)
 
 
 def predict_say(who, what):
@@ -415,7 +417,8 @@ def predict_say(who, what):
 
 def say(who, what, interact=True, *args, **kwargs):
     who = Character(who, kind=adv)
-    who(what, interact=interact, *args, **kwargs)
+    who(what, *args, interact=interact, **kwargs)
+
 
 # Used by renpy.reshow_say and extend.
 _last_say_who = None
