@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -299,6 +299,7 @@ def take_focuses():
         change_focus(global_focus, True)
 
 
+
 def focus_coordinates():
     """
     :doc: other
@@ -328,6 +329,9 @@ modal_generation = 0
 # was called.
 old_max_default = 0
 
+# The name of the old max default focus.
+old_max_default_focus_name = None
+
 def mark_modal():
     global modal_generation
     modal_generation += 1
@@ -396,11 +400,13 @@ def before_interact(roots):
             defaults.sort(key=operator.itemgetter(0))
 
         max_default, max_default_focus, max_default_screen = defaults[-1]
+        max_default_focus_name = max_default_focus.full_focus_name
 
     else:
         max_default = 0
         max_default_focus = None
         max_default_screen = None
+        max_default_focus_name = None
 
     # Should we do the max_default logic?
     should_max_default = (renpy.display.interface.last_event is None) or (renpy.display.interface.last_event.type not in [ pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP, pygame.MOUSEMOTION ])
@@ -422,7 +428,7 @@ def before_interact(roots):
     grab = replaced_by.get(id(grab), None)
 
     if override is not None:
-        d = renpy.exports.get_displayable(base=True, *override) # type: ignore
+        d = renpy.exports.get_displayable(*override, base=True) # type: ignore
 
         if (d is not None) and (current is not d) and not grab:
             current = d
@@ -460,8 +466,9 @@ def before_interact(roots):
     # If nothing has focus, focus the default if the highest priority has changed,
     # or if the default is None.
     if (should_max_default and (max_default > 0) and (current is None) and
-        (renpy.display.interface.start_interact or (max_default != old_max_default))):
+        (renpy.display.interface.start_interact or (max_default_focus_name != old_max_default_focus_name))):
 
+        explicit = True
         current = max_default_focus
         set_focused(max_default_focus, None, max_default_screen)
 
@@ -712,8 +719,17 @@ def focus_nearest(from_x0, from_y0, from_x1, from_y1,
 
     if not current:
 
+        focus_extreme(xmul, ymul, wmul, hmul)
+        current = get_focused()
+
+        if current is not None:
+            return
+
         for f in focus_list:
 
+            if f.x is False:
+                continue
+
             if not f.widget.style.keyboard_focus:
                 continue
 
@@ -760,6 +776,9 @@ def focus_nearest(from_x0, from_y0, from_x1, from_y1,
             placeless = f
             continue
 
+        if f.x is False:
+            continue
+
         if not condition(from_focus, f):
             continue
 
@@ -804,6 +823,9 @@ def focus_ordered(delta):
             placeless = f
             continue
 
+        if f.x is False:
+            continue
+
         if f.arg is not None:
             continue
 
