@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -31,7 +31,7 @@ from renpy.display.render import render, Render
 from renpy.display.matrix import Matrix2D
 
 
-class Solid(renpy.display.core.Displayable):
+class Solid(renpy.display.displayable.Displayable):
     """
     :doc: disp_imagelike
 
@@ -145,7 +145,7 @@ class Borders(object):
             )
 
 
-class Frame(renpy.display.core.Displayable):
+class Frame(renpy.display.displayable.Displayable):
     """
     :doc: disp_imagelike
     :args: (image, left=0, top=0, right=None, bottom=None, *, tile=False, **properties)
@@ -605,7 +605,7 @@ class Frame(renpy.display.core.Displayable):
         return rv
 
 
-class FileCurrentScreenshot(renpy.display.core.Displayable):
+class FileCurrentScreenshot(renpy.display.displayable.Displayable):
     """
     :doc: file_action_function
 
