@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -27,12 +27,11 @@ from typing import Any
 # This file contains displayables that move, zoom, rotate, or otherwise
 # transform displayables. (As well as displayables that support them.)
 import math
-import types # @UnresolvedImport
 
 import renpy
 from renpy.display.layout import Container
 from renpy.display.accelerator import RenderTransform
-from renpy.atl import position, any_object, bool_or_none, float_or_none, matrix, mesh
+from renpy.atl import dualangle_or_float_or_none, position, DualAngle, position_or_none, any_object, bool_or_none, float_or_none, matrix, mesh
 from renpy.display.core import absolute
 
 class Camera(renpy.object.Object):
@@ -76,10 +75,23 @@ def first_not_none(*args):
     return args[-1]
 
 
+def limit_angle(n):
+    """
+    Limits an angle to the range 0 and 360 degrees.
+    """
+
+    n = n % 360
+
+    if n < 0:
+        n += 360
+
+    return n
+
 class TransformState(renpy.object.Object):
 
     last_angle = 0.0
-    last_anchorangle = 0.0
+    last_relative_anchorangle = 0.0
+    last_absolute_anchorangle = 0.0
     last_events = True
 
     available_width = 0
@@ -113,7 +125,8 @@ class TransformState(renpy.object.Object):
             d[k] = getattr(ts, k)
 
         self.last_angle = ts.last_angle
-        self.last_anchorangle = ts.last_anchorangle
+        self.last_absolute_anchorangle = ts.last_absolute_anchorangle
+        self.last_relative_anchorangle = ts.last_relative_anchorangle
         self.last_events = ts.last_events
 
         self.available_width = ts.available_width
@@ -210,105 +223,14 @@ class TransformState(renpy.object.Object):
 
     yalign = property(get_yalign, set_yalign)
 
-    def scale(self, value, available):
+    @staticmethod
+    def scale(value, available):
         """
         Converts value to a float, scaled by the available area, if
         required.
         """
 
-        if type(value) is float:
-            return value * available
-
-        return 1.0 * value
-
-    def cartesian_to_polar_pos(self, x, y):
-        """
-        Converts cartesian coordinates to polar coordinates.
-        """
-
-        x = self.scale(x, self.available_width)
-        y = self.scale(y, self.available_height)
-
-        xaround = self.scale(self.xaround, self.available_width)
-        yaround = self.scale(self.yaround, self.available_height)
-
-        dx = x - xaround
-        dy = y - yaround
-
-        radius = math.hypot(dx, dy)
-        angle = math.atan2(dx, -dy) / math.pi * 180
-
-        if angle < 0:
-            angle += 360
-
-        return angle, radius
-
-
-    def polar_to_cartesian_pos(self, angle, radius):
-        """
-        Converts polar coordinates to cartesian coordinates.
-        """
-
-        xaround = self.scale(self.xaround, self.available_width)
-        yaround = self.scale(self.yaround, self.available_height)
-
-        angle = angle * math.pi / 180
-
-        dx = radius * math.sin(angle)
-        dy = -radius * math.cos(angle)
-
-        x = absolute(xaround + dx)
-        y = absolute(yaround + dy)
-
-        xpos = first_not_none(self.xpos, self.inherited_xpos, 0)
-        ypos = first_not_none(self.ypos, self.inherited_ypos, 0)
-
-        if type(xpos) is float:
-            x = float(x / self.available_width)
-
-        if type(ypos) is float:
-            y = float(y / self.available_height)
-
-        return x, y
-
-    def cartesian_to_polar_anchor(self, x, y):
-        """
-        Converts cartesian coordinates to polar coordinates.
-        """
-
-        xaround = self.xanchoraround
-        yaround = self.yanchoraround
-
-        dx = x - xaround
-        dy = y - yaround
-
-        radius = math.hypot(dx, dy)
-        angle = math.atan2(dx, -dy) / math.pi * 180
-
-        if angle < 0:
-            angle += 360
-
-        return angle, type(xaround)(radius)
-
-    def polar_to_cartesian_anchor(self, angle, radius):
-        """
-        Converts polar coordinates to cartesian coordinates.
-        Assumes that xanchoraround, yanchoraround, and radius
-        are all in the same units.
-        """
-
-        xaround = self.xanchoraround
-        yaround = self.yanchoraround
-
-        angle = angle * math.pi / 180
-
-        dx = radius * math.sin(angle)
-        dy = -radius * math.cos(angle)
-
-        x = type(xaround)(xaround + dx)
-        y = type(xaround)(yaround + dy)
-
-        return x, y
+        return float(absolute.compute_raw(value, available))
 
     def get_around(self):
         return (self.xaround, self.yaround)
@@ -324,86 +246,209 @@ class TransformState(renpy.object.Object):
     around = property(get_around, set_around)
     alignaround = property(get_around, set_alignaround)
 
-    def get_angle(self):
-        xpos = first_not_none(self.xpos, self.inherited_xpos, 0)
-        ypos = first_not_none(self.ypos, self.inherited_ypos, 0)
-        angle, _radius = self.cartesian_to_polar_pos(xpos, ypos)
-        return angle
+    def get_anchoraround(self):
+        return (self.xanchoraround, self.yanchoraround)
 
-    def get_radius(self):
-        xpos = first_not_none(self.xpos, self.inherited_xpos, 0)
-        ypos = first_not_none(self.ypos, self.inherited_ypos, 0)
-        _angle, radius = self.cartesian_to_polar_pos(xpos, ypos)
+    def set_anchoraround(self, value):
+        self.xanchoraround, self.yanchoraround = value
 
-        if self.radius_type is float:
-            divisor = min(self.available_width, self.available_height)
-            if divisor:
-                radius = radius / divisor
-            else:
-                radius = 0.0
+    anchoraround = property(get_anchoraround, set_anchoraround)
 
-        return self.radius_type(radius)
+    def get_pos_polar_vector(self):
+        xpos = self.scale(first_not_none(self.xpos, self.inherited_xpos, 0), self.available_width)
+        ypos = self.scale(first_not_none(self.ypos, self.inherited_ypos, 0), self.available_height)
 
-    def set_angle(self, angle):
-        self.last_angle = angle
+        xaround = self.scale(self.xaround, self.available_width)
+        yaround = self.scale(self.yaround, self.available_height)
 
-        xpos = first_not_none(self.xpos, self.inherited_xpos, 0)
-        ypos = first_not_none(self.ypos, self.inherited_ypos, 0)
-        _angle, radius = self.cartesian_to_polar_pos(xpos, ypos)
-        self.xpos, self.ypos = self.polar_to_cartesian_pos(angle, radius)
+        return (xpos - xaround, ypos - yaround)
 
-    def set_radius(self, radius):
-        self.radius_type = type(radius)
+    def get_angle(self, vector=None):
+        vector_x, vector_y = vector or self.get_pos_polar_vector()
 
-        if type(radius) is float:
-            radius = self.scale(radius, min(self.available_width, self.available_height))
+        radius = math.hypot(vector_x, vector_y)
+        angle = math.atan2(vector_x, -vector_y) / math.pi * 180
 
-        xpos = first_not_none(self.xpos, self.inherited_xpos, 0)
-        ypos = first_not_none(self.ypos, self.inherited_ypos, 0)
-        angle, old_radius = self.cartesian_to_polar_pos(xpos, ypos)
+        if angle < 0:
+            angle += 360
 
-        # Deal with the angle becoming 0.0 when the radius is 0.0.
-        if not old_radius and self.last_angle is not None:
+        if (radius == 0) and (self.last_angle is not None):
             angle = self.last_angle
 
-        self.xpos, self.ypos = self.polar_to_cartesian_pos(angle, radius)
+        return angle
+
+    def get_radius(self, vector=None):
+        vector_x, vector_y = vector or self.get_pos_polar_vector()
+
+        return absolute(math.hypot(vector_x, vector_y))
+
+    def set_angle(self, angle):
+        self.last_angle = limit_angle(angle)
+
+        radius = self.get_radius()
+
+        self.set_pos_from_angle_and_radius(angle, radius)
+
+    def set_radius(self, radius):
+        radius = self.scale(radius, min(self.available_width, self.available_height))
+
+        vector = self.get_pos_polar_vector()
+        # Deal with the angle becoming 0.0 when the radius would be 0.0.
+        if not any(vector) and (self.last_angle is not None):
+            angle = self.last_angle
+        else:
+            angle = self.get_angle(vector)
+
+        self.set_pos_from_angle_and_radius(angle, radius)
+
+    def set_pos_from_angle_and_radius(self, angle, radius):
+        xaround = self.scale(self.xaround, self.available_width)
+        yaround = self.scale(self.yaround, self.available_height)
+
+        angle = angle * math.pi / 180
+
+        dx = radius * math.sin(angle)
+        dy = -radius * math.cos(angle)
+
+        self.xpos = absolute(xaround + dx)
+        self.ypos = absolute(yaround + dy)
 
     angle = property(get_angle, set_angle)
     radius = property(get_radius, set_radius)
 
     # Anchor polar motions.
 
-    def get_anchorangle(self):
-        xanchor = first_not_none(self.xanchor, self.inherited_xanchor, 0)
-        yanchor = first_not_none(self.yanchor, self.inherited_yanchor, 0)
-        angle, _radius = self.cartesian_to_polar_anchor(xanchor, yanchor)
-        return angle
+    def get_anchor_polar_vector(self):
+        """
+        Returns a 2-tuple of 2-tuples,
+        where the first small tuple is absolute and the second tuple is relative,
+        and the first element of each tuple is in x and the second in y.
+        They represent the vector from the anchoraround point to the final anchor point.
+        """
+        xanchoraround = position.from_any(self.xanchoraround)
+        yanchoraround = position.from_any(self.yanchoraround)
+        xanchor = position.from_any(first_not_none(self.xanchor, self.inherited_xanchor, 0))
+        yanchor = position.from_any(first_not_none(self.yanchor, self.inherited_yanchor, 0))
 
-    def get_anchorradius(self):
-        xanchor = first_not_none(self.xanchor, self.inherited_xanchor, 0)
-        yanchor = first_not_none(self.yanchor, self.inherited_yanchor, 0)
-        _angle, radius = self.cartesian_to_polar_anchor(xanchor, yanchor)
+        absolute_vector = (xanchor.absolute-xanchoraround.absolute, yanchor.absolute-yanchoraround.absolute)
+        relative_vector = (xanchor.relative-xanchoraround.relative, yanchor.relative-yanchoraround.relative)
 
-        return self.radius_type(radius)
+        return absolute_vector, relative_vector
+
+    def get_anchorangle(self, polar_vectors=None):
+        """
+        Returns a DualAngle object, from the oriented angle in degrees, with 0 as the top direction and 90 as the right,
+        of the vector going from (xanchoraround, yanchoraround) to (xanchor, yanchor).
+        The absolute part of the angle is the angle between the absolute parts of the vectors,
+        and the relative part, of the relative parts.
+        """
+        (absolute_vector_x, absolute_vector_y), (relative_vector_x, relative_vector_y) = polar_vectors or self.get_anchor_polar_vector()
+
+        absolute_radius = math.hypot(absolute_vector_x, absolute_vector_y)
+        relative_radius = math.hypot(relative_vector_x, relative_vector_y)
+        absolute_angle = math.atan2(absolute_vector_x, -absolute_vector_y) / math.pi * 180
+        relative_angle = math.atan2(relative_vector_x, -relative_vector_y) / math.pi * 180
+        if absolute_angle < 0:
+            absolute_angle += 360
+        if relative_angle < 0:
+            relative_angle += 360
+
+        if (absolute_radius == 0) and (self.last_absolute_anchorangle is not None):
+            absolute_angle = self.last_absolute_anchorangle
+        if (relative_radius == 0) and (self.last_relative_anchorangle is not None):
+            relative_angle = self.last_relative_anchorangle
+
+        return DualAngle(absolute_angle, relative_angle)
+
+    def get_anchorradius(self, polar_vectors=None):
+        """
+        Returns the distance between (xanchoraround, yanchoraround) and (xanchor, yanchor),
+        as a position object.
+        """
+        (absolute_vector_x, absolute_vector_y), (relative_vector_x, relative_vector_y) = polar_vectors or self.get_anchor_polar_vector()
+
+        return position(
+            absolute=math.hypot(absolute_vector_x, absolute_vector_y), # type: ignore
+            relative=math.hypot(relative_vector_x, relative_vector_y),
+        )
 
     def set_anchorangle(self, angle):
-        self.last_anchorangle = angle
+        """
+        Computes the anchorradius (as a position object),
+        and set xanchor and yanchor such that the anchorradius (both the absolute and relative parts)
+        remain the same, and the anchorangle (as explained above) is the given one.
+        """
+        if isinstance(angle, DualAngle):
+            absolute_anchorangle = angle.absolute
+            relative_anchorangle = angle.relative
+        else:
+            absolute_anchorangle = relative_anchorangle = angle
 
-        xanchor = first_not_none(self.xanchor, self.inherited_xanchor, 0)
-        yanchor = first_not_none(self.yanchor, self.inherited_yanchor, 0)
-        _angle, radius = self.cartesian_to_polar_anchor(xanchor, yanchor)
-        self.xanchor, self.yanchor = self.polar_to_cartesian_anchor(angle, radius)
+        self.last_absolute_anchorangle = limit_angle(absolute_anchorangle)
+        self.last_relative_anchorangle = limit_angle(relative_anchorangle)
 
-    def set_anchorradius(self, radius):
-        xanchor = first_not_none(self.xanchor, self.inherited_xanchor, 0)
-        yanchor = first_not_none(self.yanchor, self.inherited_yanchor, 0)
-        angle, old_radius = self.cartesian_to_polar_anchor(xanchor, yanchor)
+        anchorradius = self.anchorradius
 
-        # Deal with the angle becoming 0.0 when the radius is 0.0.
-        if not old_radius and self.last_anchorangle is not None:
-            angle = self.last_anchorangle
+        self.set_anchor_from_anchorangle_and_anchorradius(
+            absolute_anchorangle,
+            relative_anchorangle,
+            anchorradius.absolute,
+            anchorradius.relative,
+        )
 
-        self.xanchor, self.yanchor = self.polar_to_cartesian_anchor(angle, radius)
+    def set_anchorradius(self, anchorradius):
+        """
+        Computes the anchorangle (as a DualAngle object),
+        and set xanchor and yanchor such that the anchorangle stays the same,
+        and the anchorradius (as explained above) is the given one.
+        """
+        anchorradius = position.from_any(anchorradius)
+
+        polar_vectors = self.get_anchor_polar_vector()
+        anchorangle = self.get_anchorangle(polar_vectors)
+        old_anchorradius = self.get_anchorradius(polar_vectors)
+
+        absolute_anchorangle = anchorangle.absolute
+        relative_anchorangle = anchorangle.relative
+
+        if (not old_anchorradius.absolute) and (self.last_absolute_anchorangle is not None):
+            absolute_anchorangle = self.last_absolute_anchorangle
+        if (not old_anchorradius.relative) and (self.last_relative_anchorangle is not None):
+            relative_anchorangle = self.last_relative_anchorangle
+
+        self.set_anchor_from_anchorangle_and_anchorradius(
+            absolute_anchorangle,
+            relative_anchorangle,
+            anchorradius.absolute,
+            anchorradius.relative,
+        )
+
+    def set_anchor_from_anchorangle_and_anchorradius(self,
+            absolute_anchorangle,
+            relative_anchorangle,
+            absolute_anchorradius,
+            relative_anchorradius,
+        ):
+
+        xanchoraround = position.from_any(self.xanchoraround)
+        yanchoraround = position.from_any(self.yanchoraround)
+
+        absolute_anchorangle = absolute_anchorangle * math.pi / 180
+        relative_anchorangle = relative_anchorangle * math.pi / 180
+
+        absolute_dx = absolute_anchorradius * math.sin(absolute_anchorangle)
+        absolute_dy = -absolute_anchorradius * math.cos(absolute_anchorangle)
+        relative_dx = relative_anchorradius * math.sin(relative_anchorangle)
+        relative_dy = -relative_anchorradius * math.cos(relative_anchorangle)
+
+        self.xanchor = position(
+            absolute=xanchoraround.absolute + absolute_dx,
+            relative=xanchoraround.relative + relative_dx,
+        )
+        self.yanchor = position(
+            absolute=yanchoraround.absolute + absolute_dy,
+            relative=yanchoraround.relative + relative_dy,
+        )
 
     anchorangle = property(get_anchorangle, set_anchorangle)
     anchorradius = property(get_anchorradius, set_anchorradius)
@@ -424,14 +469,11 @@ class TransformState(renpy.object.Object):
 
     anchor = property(get_anchor, set_anchor)
 
-    def get_align(self):
-        return self.xpos, self.ypos
-
     def set_align(self, value):
         self.xanchor, self.yanchor = value
         self.xpos, self.ypos = value
 
-    align = property(get_align, set_align)
+    align = property(get_pos, set_align)
 
     def get_offset(self):
         return self.xoffset, self.yoffset
@@ -463,28 +505,25 @@ class TransformState(renpy.object.Object):
         self.xpos = value
         self.xanchor = 0.5
 
-    def get_xcenter(self):
+    def get_xpos(self):
         return self.xpos
 
     def set_ycenter(self, value):
         self.ypos = value
         self.yanchor = 0.5
 
-    def get_ycenter(self):
+    def get_ypos(self):
         return self.ypos
 
-    xcenter = property(get_xcenter, set_xcenter)
-    ycenter = property(get_ycenter, set_ycenter)
-
-    def get_xycenter(self):
-        return self.xcenter, self.ycenter
+    xcenter = property(get_xpos, set_xcenter)
+    ycenter = property(get_ypos, set_ycenter)
 
     def set_xycenter(self, value):
         if value is None:
             value = (None, None)
         self.xcenter, self.ycenter = value
 
-    xycenter = property(get_xycenter, set_xycenter)
+    xycenter = property(get_pos, set_xycenter)
 
 
 class Proxy(object):
@@ -496,7 +535,16 @@ class Proxy(object):
         self.name = name
 
     def __get__(self, instance, owner):
-        return getattr(instance.state, self.name)
+
+        def simplify_position(v):
+            if isinstance(v, tuple):
+                return tuple(simplify_position(i) for i in v)
+            elif isinstance(v, position):
+                return v.simplify()
+            else:
+                return v
+
+        return simplify_position(getattr(instance.state, self.name))
 
     def __set__(self, instance, value):
         return setattr(instance.state, self.name, value)
@@ -1007,10 +1055,8 @@ class Transform(Container):
                 cw, ch = self.child_size
                 rw, rh = self.render_size
 
-                if xanchor.__class__ is float:
-                    xanchor *= cw
-                if yanchor.__class__ is float:
-                    yanchor *= ch
+                xanchor = absolute.compute_raw(xanchor, cw)
+                yanchor = absolute.compute_raw(yanchor, ch)
 
                 xanchor -= cw / 2.0
                 yanchor -= ch / 2.0
@@ -1020,8 +1066,8 @@ class Transform(Container):
                 xanchor += rw / 2.0
                 yanchor += rh / 2.0
 
-                xanchor = renpy.display.core.absolute(xanchor)
-                yanchor = renpy.display.core.absolute(yanchor)
+                xanchor = absolute(xanchor)
+                yanchor = absolute(yanchor)
 
                 rv = (xpos, ypos, xanchor, yanchor, xoffset, yoffset, subpixel)
 
@@ -1166,15 +1212,15 @@ add_property("additive", float, 0.0)
 add_property("alpha", float, 1.0)
 add_property("blend", any_object, None)
 add_property("blur", float_or_none, None)
-add_property("corner1", (position, position), None)
-add_property("corner2", (position, position), None)
-add_property("crop", (position, position, position, position), None)
+add_property("corner1", (position_or_none, position_or_none), None)
+add_property("corner2", (position_or_none, position_or_none), None)
+add_property("crop", (position_or_none, position_or_none, position_or_none, position_or_none), None)
 add_property("crop_relative", bool_or_none, None)
 add_property("debug", any_object, None)
 add_property("delay", float, 0)
 add_property("events", bool, True)
 add_property("fit", str, None)
-add_property("matrixanchor", (position, position), None)
+add_property("matrixanchor", (position_or_none, position_or_none), None)
 add_property("matrixcolor", matrix, None)
 add_property("matrixtransform", matrix, None)
 add_property("maxsize", (int, int), None)
@@ -1195,23 +1241,23 @@ add_property("subpixel", bool, False)
 add_property("transform_anchor", bool, False)
 add_property("zoom", float, 1.0)
 
-add_property("xanchoraround", float, 0.5)
-add_property("xanchor", position, None, diff=4)
-add_property("xaround", position, 0.0)
-add_property("xoffset", float, 0.0)
+add_property("xanchoraround", position_or_none, 0.5)
+add_property("xanchor", position_or_none, None, diff=4)
+add_property("xaround", position_or_none, 0.0)
+add_property("xoffset", absolute, 0.0)
 add_property("xpan", float_or_none, None)
-add_property("xpos", position, None, diff=4)
-add_property("xsize", position, None)
+add_property("xpos", position_or_none, None, diff=4)
+add_property("xsize", position_or_none, None)
 add_property("xtile", int, 1)
 add_property("xzoom", float, 1.0)
 
-add_property("yanchoraround", float, 0.5)
-add_property("yanchor", position, None, diff=4)
-add_property("yaround", position, 0.0)
-add_property("yoffset", float, 0.0)
+add_property("yanchoraround", position_or_none, 0.5)
+add_property("yanchor", position_or_none, None, diff=4)
+add_property("yaround", position_or_none, 0.0)
+add_property("yoffset", absolute, 0.0)
 add_property("ypan", float_or_none, None)
-add_property("ypos", position, None, diff=4)
-add_property("ysize", position, None)
+add_property("ypos", position_or_none, None, diff=4)
+add_property("ysize", position_or_none, None)
 add_property("ytile", int, 1)
 add_property("yzoom", float, 1.0)
 
@@ -1230,22 +1276,23 @@ add_gl_property("gl_texture_wrap")
 
 ALIASES = {
     "alignaround" : (float, float),
-    "align" : (float, float),
-    "anchor" : (position, position),
-    "anchorangle" : float,
-    "anchorradius" : position,
+    "align" : (position_or_none, position_or_none), # document as (float, float)
+    "anchor" : (position_or_none, position_or_none),
+    "anchorangle" : dualangle_or_float_or_none,
+    "anchoraround" : (position_or_none, position_or_none),
+    "anchorradius" : position_or_none,
     "angle" : float,
-    "around" : (position, position),
-    "offset" : (int, int),
-    "pos" : (position, position),
-    "radius" : position,
+    "around" : (position_or_none, position_or_none),
+    "offset" : (absolute, absolute),
+    "pos" : (position_or_none, position_or_none),
+    "radius" : position_or_none,
     "size" : (int, int),
-    "xalign" : float,
-    "xcenter" : position,
-    "xycenter" : (position, position),
-    "xysize" : (position, position),
-    "yalign" : float,
-    "ycenter" : position,
+    "xalign" : position_or_none, # document as float,
+    "xcenter" : position_or_none,
+    "xycenter" : (position_or_none, position_or_none),
+    "xysize" : (position_or_none, position_or_none),
+    "yalign" : position_or_none, # document as float
+    "ycenter" : position_or_none,
     }
 
 renpy.atl.PROPERTIES.update(ALIASES)
