@@ -1,4 +1,4 @@
-# Copyright 2004-2023 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -249,33 +249,43 @@ def get_movie_texture_web(channel, mask_channel, side_mask, mipmap):
     return tex, new
 
 
-def render_movie(channel, width, height, group=None):
-    tex, _new = get_movie_texture(channel)
+def resize_movie(r, width, height):
+    """
+    A utility function to resize a Render or texture to the given
+    dimensions.
+    """
 
-    if group is not None:
-        if tex is None:
-            tex = group_texture.get(group, None)
-        else:
-            group_texture[group] = tex
-
-    if tex is None:
+    if r is None:
         return None
 
-    sw, sh = tex.get_size()
+    rv = renpy.display.render.Render(width, height)
 
-    scale = min(1.0 * width / sw, 1.0 * height / sh)
+    sw, sh = r.get_size()
+
+    if not (sw and sh):
+        return rv
+
+    scale = min(1.0 * width / sw, 1.0 * height / sh) # type: float
 
     dw = scale * sw
     dh = scale * sh
 
-    rv = renpy.display.render.Render(width, height)
     rv.forward = renpy.display.matrix.Matrix2D(1.0 / scale, 0.0, 0.0, 1.0 / scale)
     rv.reverse = renpy.display.matrix.Matrix2D(scale, 0.0, 0.0, scale)
-    rv.blit(tex, (int((width - dw) / 2), int((height - dh) / 2)))
+    rv.blit(r, (int((width - dw) / 2), int((height - dh) / 2)))
 
     return rv
 
 
+def render_movie(channel, width, height):
+    """
+    Called from the Draw objects to render and scale a fullscreen movie.
+    """
+
+    tex, _new = get_movie_texture(channel)
+    return resize_movie(tex, width, height)
+
+
 def default_play_callback(old, new): # @UnusedVariable
 
     renpy.audio.music.play(new._play, channel=new.channel, loop=new.loop, synchro_start=True)
@@ -286,17 +296,13 @@ def default_play_callback(old, new): # @UnusedVariable
 # A serial number that's used to generated movie channels.
 movie_channel_serial = 0
 
-class Movie(renpy.display.core.Displayable):
+class Movie(renpy.display.displayable.Displayable):
     """
     :doc: movie
+    :args: (*, size=None, channel="movie", play=None, side_mask=False, mask=None, mask_channel=None, start_image=None, image=None, play_callback=None, loop=True, group=None, **properties)
 
     This is a displayable that shows the current movie.
 
-    `fps`
-        The framerate that the movie should be shown at. (This is currently
-        ignored, but the parameter is kept for backwards compatibility.
-        The framerate is auto-detected.)
-
     `size`
         This should be specified as either a tuple giving the width and
         height of the movie, or None to automatically adjust to the size
@@ -348,7 +354,8 @@ class Movie(renpy.display.core.Displayable):
         to create a slimmed-down mobile version that does not use movie
         sprites.) Users can also choose to fall back to this image as a
         preference if video is too taxing for their system. The image will
-        also be used if the video plays, and then the movie ends.
+        also be used if the video plays, and then the movie ends, unless
+        `group` is given.
 
     `play_callback`
         If not None, a function that's used to start the movies playing.
@@ -386,6 +393,11 @@ class Movie(renpy.display.core.Displayable):
         the previous frame, the last frame from that movie will be used for
         this movie. This can prevent flashes of transparency when switching
         between two movies.
+
+    `keep_last_frame`
+        If true, and the movie has ended, the last frame will be displayed,
+        rather than the movie being hidden. This only works if `loop` is
+        false. (This behavior will also occur if `group` is set.)
     """
 
     fullscreen = False
@@ -456,7 +468,9 @@ class Movie(renpy.display.core.Displayable):
         self.ensure_channel(self.channel)
         self.ensure_channel(self.mask_channel)
 
-    def __init__(self, fps=24, size=None, channel="movie", play=None, mask=None, mask_channel=None, image=None, play_callback=None, side_mask=False, loop=True, start_image=None, group=None, **properties):
+    keep_last_frame_serial = 0
+
+    def __init__(self, fps=24, size=None, channel="movie", play=None, mask=None, mask_channel=None, image=None, play_callback=None, side_mask=False, loop=True, start_image=None, group=None, keep_last_frame=False, **properties):
 
         global movie_channel_serial
 
@@ -497,10 +511,11 @@ class Movie(renpy.display.core.Displayable):
 
         self.play_callback = play_callback
 
-        self.group = group
+        if group is None and keep_last_frame:
+            group = "_keep_last_frame_" + str(Movie.keep_last_frame_serial)
+            Movie.keep_last_frame_serial += 1
 
-        if (self.channel == "movie") and (renpy.config.hw_video) and renpy.mobile:
-            raise Exception("Movie(channel='movie') doesn't work on mobile when config.hw_video is true. (Use a different channel argument.)")
+        self.group = group
 
     def _handles_event(self, event):
         return event == "show"
@@ -535,42 +550,31 @@ class Movie(renpy.display.core.Displayable):
 
             return rv
 
-        if self.size is None:
-
-            tex, _ = get_movie_texture(self.channel, self.mask_channel, self.side_mask, self.style.mipmap)
-
-            if self.group is not None:
-                if tex is None:
-                    tex = group_texture.get(self.group, None)
-                else:
-                    group_texture[self.group] = tex
-
-            if (not not_playing) and (tex is not None):
-                width, height = tex.get_size()
-
-                rv = renpy.display.render.Render(width, height)
-                rv.blit(tex, (0, 0))
-
-            elif (not not_playing) and (self.start_image is not None):
-                surf = renpy.display.render.render(self.start_image, width, height, st, at)
-                w, h = surf.get_size()
-                rv = renpy.display.render.Render(w, h)
-                rv.blit(surf, (0, 0))
+        tex, _ = get_movie_texture(self.channel, self.mask_channel, self.side_mask, self.style.mipmap)
 
+        if self.group is not None:
+            if tex is None:
+                tex = group_texture.get(self.group, None)
             else:
-                rv = renpy.display.render.Render(0, 0)
+                group_texture[self.group] = tex
+
+        if (not not_playing) and (tex is not None):
+            width, height = tex.get_size()
+
+            rv = renpy.display.render.Render(width, height)
+            rv.blit(tex, (0, 0))
+
+        elif (not not_playing) and (self.start_image is not None):
+            surf = renpy.display.render.render(self.start_image, width, height, st, at)
+            w, h = surf.get_size()
+            rv = renpy.display.render.Render(w, h)
+            rv.blit(surf, (0, 0))
 
         else:
+            rv = renpy.display.render.Render(0, 0)
 
-            w, h = self.size
-
-            if not playing:
-                rv = None
-            else:
-                rv = render_movie(self.channel, w, h, group=self.group)
-
-            if rv is None:
-                rv = renpy.display.render.Render(w, h)
+        if self.size is not None:
+            rv = resize_movie(rv, self.size[0], self.size[1])
 
         # Usually we get redrawn when the frame is ready - but we want
         # the movie to disappear if it's ended, or if it hasn't started
@@ -681,8 +685,17 @@ def frequent():
             if m.group is not None:
                 group_texture[m.group] = old_group_texture.get(m.group, None)
 
+    if fullscreen:
+
+            c = renpy.audio.audio.get_channel("movie")
+
+            if c.video_ready():
+                return True
+            else:
+                return False
+
     # Determine if we need to redraw.
-    if displayable_channels:
+    elif displayable_channels:
 
         update = True
 
@@ -707,13 +720,5 @@ def frequent():
 
         return False
 
-    elif fullscreen and not ((renpy.android or renpy.ios) and renpy.config.hw_video):
-
-        c = renpy.audio.audio.get_channel("movie")
-
-        if c.video_ready():
-            return True
-        else:
-            return False
 
     return False
