@@ -1,4 +1,4 @@
-# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -30,12 +30,11 @@ import time
 import zipfile
 import gc
 import linecache
+import json
 
 import renpy
 import renpy.game as game
 
-import __main__
-
 last_clock = time.time()
 
 
@@ -151,9 +150,10 @@ def load_rpe(fn):
     with zipfile.ZipFile(fn) as zfn:
         autorun = zfn.read("autorun.py")
 
+    if fn in sys.path:
+        sys.path.remove(fn)
     sys.path.insert(0, fn)
-    exec(autorun, dict())
-
+    exec(autorun, {'__file__': os.path.join(fn, "autorun.py")})
 
 def choose_variants():
 
@@ -253,8 +253,9 @@ def choose_variants():
         renpy.config.variants.insert(0, 'web') # type: ignore
 
         # mobile
-        userAgent = emscripten.run_script_string(r'''navigator.userAgent''')
-        mobile = re.search('Mobile|Android|iPad|iPhone', userAgent)
+        mobile = emscripten.run_script_int(
+            r'''/Mobile|Android|iPad|iPhone/.test(navigator.userAgent)
+            || (navigator.userAgent.indexOf("Mac") != -1 && navigator.maxTouchPoints > 1)''')
         if mobile:
             renpy.config.variants.insert(0, 'mobile') # type: ignore
         # Reserve android/ios for when the OS API is exposed
@@ -300,49 +301,23 @@ def choose_variants():
         renpy.config.variants.insert(0, 'large') # type: ignore
 
 
-def android_searchpath():
+def load_build_info():
     """
-    Determines the searchpath on Android.
+    Loads cache/build_info.json, and uses it to initialize the
+    renpy.game.build_info dictionary.
     """
 
-    # The default gamedir, in private.
-    renpy.config.searchpath = [ renpy.config.gamedir ]
-
-    # The public android directory.
-    if "ANDROID_PUBLIC" in os.environ:
-        android_game = os.path.join(os.environ["ANDROID_PUBLIC"], "game")
-
-        if os.path.exists(android_game):
-            renpy.config.searchpath.insert(0, android_game)
-
-    # Asset packs.
-    packs = [
-        "ANDROID_PACK_FF1",
-        "ANDROID_PACK_FF2",
-        "ANDROID_PACK_FF3",
-        "ANDROID_PACK_FF4",
-    ]
-
-    for i in packs:
-        if i not in os.environ:
-            continue
-
-        assets = os.environ[i]
-
-        for i in [ "renpy/common", "game" ]:
-            dn = os.path.join(assets, i)
-            if os.path.isdir(dn):
-                renpy.config.searchpath.append(dn)
-
-    print("Android search paths:" , " ".join(renpy.config.searchpath))
+    try:
+        f = renpy.exports.open_file("cache/build_info.json", "utf-8")
+        renpy.game.build_info = json.load(f)
+    except Exception:
+        renpy.game.build_info = { "info" : { } }
 
 
 def main():
 
     gc.set_threshold(*renpy.config.gc_thresholds)
 
-    log_clock("Bootstrap to the start of init.init")
-
     renpy.game.exception_info = 'Before loading the script.'
 
     # Clear the line cache, since the script may have changed.
@@ -367,33 +342,21 @@ def main():
     choose_variants()
     renpy.display.touch = "touch" in renpy.config.variants
 
-    log_clock("Early init")
+    if (renpy.android or renpy.ios) and not renpy.config.log_to_stdout:
+        print("Version:", renpy.version)
 
     # Note the game directory.
     game.basepath = renpy.config.gamedir
-    renpy.config.searchpath = [ renpy.config.gamedir ]
-
-    # Find the common directory.
-    commondir = __main__.path_to_common(renpy.config.renpy_base) # E1101 @UndefinedVariable
-
-    if os.path.isdir(commondir):
-        renpy.config.searchpath.append(commondir)
-        renpy.config.commondir = commondir
-    else:
-        renpy.config.commondir = None
-
-    # Add path from env variable, if any
-    if "RENPY_SEARCHPATH" in os.environ:
-        renpy.config.searchpath.extend(os.environ["RENPY_SEARCHPATH"].split("::"))
-
-    if renpy.android:
-        renpy.config.commondir = None
-
-        android_searchpath()
+    renpy.config.commondir = renpy.__main__.path_to_common(renpy.config.renpy_base) # E1101 @UndefinedVariable
+    renpy.config.searchpath = renpy.__main__.predefined_searchpath(renpy.config.commondir) # E1101 @UndefinedVariable
 
     # Load Ren'Py extensions.
     for dir in renpy.config.searchpath: # @ReservedAssignment
-        for fn in os.listdir(dir):
+
+        if not os.path.isdir(dir):
+            continue
+
+        for fn in sorted(os.listdir(dir)):
             if fn.lower().endswith(".rpe"):
                 load_rpe(dir + "/" + fn)
 
@@ -427,7 +390,9 @@ def main():
     # Start auto-loading.
     renpy.loader.auto_init()
 
-    log_clock("Loader init")
+    load_build_info()
+
+    log_clock("Early init")
 
     # Initialize the log.
     game.log = renpy.python.RollbackLog()
@@ -473,7 +438,7 @@ def main():
             if dn is None:
                 continue
 
-            if not os.path.isfile(os.path.join(dn, fn + ".rpy")):
+            if not os.path.isfile(os.path.join(dn, fn + ".rpy")) and not os.path.isfile(os.path.join(dn, fn + "_ren.py")):
 
                 try:
                     name = os.path.join(dn, fn + ".rpyc")
@@ -505,11 +470,14 @@ def main():
 
     # Find the save directory.
     if renpy.config.savedir is None:
-        renpy.config.savedir = __main__.path_to_saves(renpy.config.gamedir) # E1101 @UndefinedVariable
+        renpy.config.savedir = renpy.__main__.path_to_saves(renpy.config.gamedir) # E1101 @UndefinedVariable
 
     if renpy.game.args.savedir: # type: ignore
         renpy.config.savedir = renpy.game.args.savedir # type: ignore
 
+    # Init the save token system.
+    renpy.savetoken.init()
+
     # Init preferences.
     game.persistent = renpy.persistent.init()
     game.preferences = game.persistent._preferences
@@ -524,13 +492,11 @@ def main():
     # Init save locations and loadsave.
     renpy.savelocation.init()
 
-    # We need to be 100% sure we kill the savelocation thread.
     try:
-
-        # Init save slots.
+        # Init save slots and save tokens.
         renpy.loadsave.init()
-
-        log_clock("Loading save slot metadata.")
+        renpy.savetoken.upgrade_all_savefiles()
+        log_clock("Loading save slot metadata")
 
         # Load persistent data from all save locations.
         renpy.persistent.update()
@@ -550,7 +516,9 @@ def main():
 
         renpy.game.exception_info = 'While executing init code:'
 
-        for _prio, node in game.script.initcode:
+        for id_, (_prio, node) in enumerate(game.script.initcode):
+
+            renpy.game.initcode_ast_id = id_
 
             if isinstance(node, renpy.ast.Node):
                 node_start = time.time()
@@ -578,6 +546,8 @@ def main():
         for i in renpy.game.post_init:
             i()
 
+        renpy.config.post_init()
+
         renpy.game.script.report_duplicate_labels()
 
         # Sort the images.
@@ -595,6 +565,10 @@ def main():
         renpy.atl.compile_all()
         log_clock("Analyze and compile ATL")
 
+        renpy.savelocation.init()
+        renpy.loadsave.init()
+        log_clock("Reloading save slot metadata")
+
         # Index the archive files. We should not have loaded an image
         # before this point. (As pygame will not have been initialized.)
         # We need to do this again because the list of known archives
@@ -609,7 +583,7 @@ def main():
 
         renpy.dump.dump(False)
         renpy.game.script.make_backups()
-        log_clock("Dump and make backups.")
+        log_clock("Dump and make backups")
 
         # Initialize image cache.
         renpy.display.im.cache.init()
@@ -619,6 +593,9 @@ def main():
         renpy.python.make_clean_stores()
         log_clock("Making clean stores")
 
+        # Init the keymap.
+        renpy.display.behavior.init_keymap()
+
         gc.collect(2)
 
         if gc.garbage:
@@ -637,7 +614,7 @@ def main():
         else:
             gc.set_threshold(700, 10, 10)
 
-        log_clock("Initial gc.")
+        log_clock("Initial gc")
 
         # Start debugging file opens.
         renpy.debug.init_main_thread_open()
@@ -660,14 +637,20 @@ def main():
                     run(restart)
                 finally:
                     restart = (renpy.config.end_game_transition, "_invoke_main_menu", "_main_menu")
-                    renpy.persistent.update(True)
-                    renpy.persistent.save_MP()
+
+            except renpy.game.QuitException:
+
+                renpy.audio.audio.fadeout_all()
+                raise
 
             except game.FullRestartException as e:
                 restart = e.reason
 
             finally:
 
+                renpy.persistent.update(True)
+                renpy.persistent.save_on_quit_MP()
+
                 # Reset live2d if it exists.
                 try:
                     renpy.gl2.live2d.reset_states()
