@@ -1,4 +1,4 @@
-# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -36,7 +36,6 @@ from renpy.compat.pickle import dump, dumps, loads
 
 # The class that's used to hold the persistent data.
 
-
 class Persistent(object):
 
     def __init__(self):
@@ -55,6 +54,9 @@ class Persistent(object):
 
         return None
 
+    def _hasattr(self, field_name):
+        return field_name in self.__dict__
+
     def _clear(self, progress=False):
         """
         Resets the persistent data.
@@ -77,6 +79,8 @@ class Persistent(object):
             self._chosen.clear()
             self._seen_audio.clear()
 
+        renpy.exports.execute_default_statement()
+
     def _update(self):
         """
         Updates the persistent data to be the latest version of
@@ -199,11 +203,15 @@ def load(filename):
     # Unserialize the persistent data.
     try:
         with open(filename, "rb") as f:
-            s = zlib.decompress(f.read())
-        persistent = loads(s)
-    except Exception:
-        import renpy.display
+            do = zlib.decompressobj()
+            s = do.decompress(f.read())
 
+            if not renpy.savetoken.check_persistent(s, do.unused_data.decode("utf-8")):
+                return None
+
+        persistent = loads(s)
+
+    except Exception:
         try:
             renpy.display.log.write("Loading persistent.")
             renpy.display.log.exception()
@@ -225,6 +233,9 @@ def init():
     disk, so that we can configure the savelocation system.
     """
 
+    if renpy.config.early_developer and not PY2:
+        init_debug_pickler()
+
     filename = os.path.join(renpy.config.savedir, "persistent.new") # type: ignore
     persistent = load(filename)
 
@@ -236,14 +247,40 @@ def init():
         persistent = Persistent()
 
     # Create the backup of the persistent data.
-    v = vars(persistent)
-
-    for k, v in vars(persistent).items():
+    for k, v in persistent.__dict__.items():
         backup[k] = safe_deepcopy(v)
 
     return persistent
 
 
+def init_debug_pickler():
+    import io, pickle
+
+    safe_types = set()
+
+    for d in renpy.python.store_dicts.values():
+        for v in d.values():
+            if isinstance(v, type):
+                safe_types.add(v)
+
+    class DebugPickler(pickle.Pickler):
+        def reducer_override(self, obj):
+            t = obj if isinstance(obj, type) else type(obj)
+
+            if t not in safe_types and t.__module__.startswith("store"):
+                cls = (t.__module__ + '.' + t.__qualname__)[6:]
+                raise TypeError("{} is not safe for use in persistent.".format(cls))
+
+            return NotImplemented # lets normal reducing take place
+
+    global dumps
+
+    def dumps(o):
+        b = io.BytesIO()
+        DebugPickler(b, renpy.compat.pickle.PROTOCOL).dump(o)
+        return b.getvalue()
+
+
 # A map from field name to merge function.
 registry = { }
 
@@ -306,8 +343,8 @@ def merge(other):
 
     persistent = renpy.game.persistent
 
-    pvars = vars(persistent)
-    ovars = vars(other)
+    pvars = persistent.__dict__
+    ovars = other.__dict__
 
     fields = set(pvars.keys()) | set(ovars.keys())
 
@@ -405,12 +442,17 @@ def save():
     Saves the persistent data to disk.
     """
 
+    if not renpy.config.save_persistent:
+        return
+
     if not should_save_persistent:
         return
 
     try:
-        data = zlib.compress(dumps(renpy.game.persistent), 3)
-        renpy.loadsave.location.save_persistent(data)
+        data = dumps(renpy.game.persistent)
+        compressed = zlib.compress(data, 3)
+        compressed += renpy.savetoken.sign_data(data).encode("utf-8")
+        renpy.loadsave.location.save_persistent(compressed)
     except Exception:
         if renpy.config.developer:
             raise
@@ -418,13 +460,8 @@ def save():
     global persistent_mtime
 
     # Prevent updates just after save
-    mtime = persistent_mtime
-
     for mtime, _data in renpy.loadsave.location.load_persistent():
-        if mtime <= persistent_mtime:
-            continue
-
-    persistent_mtime = mtime
+        persistent_mtime = max(persistent_mtime, mtime)
 
 
 
@@ -432,55 +469,90 @@ def save():
 # MultiPersistent
 ################################################################################
 
-
-save_MP_instances = weakref.WeakSet()
+# `_MultiPersistent` instances from `MultiPersistent` calls.
+MP_instances = weakref.WeakSet()
 
 
 def save_MP():
-    for ins in save_MP_instances:
-        ins.save()
+    """
+    Called `save` for each `_MultiPersistent` instance.
+    """
+    for instance in MP_instances:
+        instance.save()
 
 
+def save_on_quit_MP():
+    """
+    Called `save` for each `_MultiPersistent` instance to be saved on exit.
+    """
+    for instance in MP_instances:
+        if instance._save_on_quit:
+            instance.save()
+
+
+def get_MP(name):
+    """
+    Returns `_MultiPersistent` instance if exists.
+    """
+    for instance in MP_instances:
+        if instance._name == name:
+            return instance
+
 class _MultiPersistent(object):
 
     _filename = ""
+    _name = ""
+    _save_on_quit = False
 
     def __getstate__(self):
         state = self.__dict__.copy()
         del state['_filename']
+        del state['_name']
+        del state['_save_on_quit']
         return state
 
     def __setstate__(self, state):
         self.__dict__.update(state)
 
     def __getattr__(self, name):
-
         if name.startswith("__") and name.endswith("__"):
             raise AttributeError()
 
         return None
 
     def save(self):
-
-        fn = self._filename
-        with open(fn + ".new", "wb") as f:
-            dump(self, f)
-
         try:
-            os.rename(fn + ".new", fn)
-        except Exception:
-            os.unlink(fn)
-            os.rename(fn + ".new", fn)
+            fn = self._filename
+            with open(fn + ".new", "wb") as f:
+                dump(self, f)
+        except OSError as e:
+            if renpy.config.developer:
+                raise e
+        else:
+            try:
+                os.rename(fn + ".new", fn)
+            except Exception:
+                os.unlink(fn)
+                os.rename(fn + ".new", fn)
 
 
 def MultiPersistent(name, save_on_quit=False):
-
-    name = renpy.exports.fsdecode(name)
+    """
+    Returns `_MultiPersistent` object.
+    """
 
     if not renpy.game.context().init_phase:
         raise Exception("MultiPersistent objects must be created during the init phase.")
 
-    if renpy.android or renpy.ios:
+    name = renpy.exports.fsdecode(name)
+    rv = get_MP(name)
+    if rv is not None:
+        return rv
+
+    if "RENPY_MULTIPERSISTENT" in os.environ:
+        files = [ renpy.exports.fsdecode(os.environ["RENPY_MULTIPERSISTENT"]) ]
+
+    elif renpy.android or renpy.ios:
         # Due to the security policy of mobile devices, we store MultiPersistent
         # in the same place as common persistent.
         # This is better than not working at all.
@@ -490,7 +562,11 @@ def MultiPersistent(name, save_on_quit=False):
         files = [ os.path.expanduser("~/RenPy/Persistent") ]
 
         if 'APPDATA' in os.environ:
-            files.append(renpy.exports.fsdecode(os.environ['APPDATA']) + "/RenPy/persistent")
+            files.append(
+                os.path.join(
+                    renpy.exports.fsdecode(os.environ['APPDATA']), "RenPy", "persistent"
+                )
+            )
 
     elif renpy.macintosh:
         files = [ os.path.expanduser("~/.renpy/persistent"),
@@ -498,16 +574,12 @@ def MultiPersistent(name, save_on_quit=False):
     else:
         files = [ os.path.expanduser("~/.renpy/persistent") ]
 
-    if "RENPY_MULTIPERSISTENT" in os.environ:
-        files = [ renpy.exports.fsdecode(os.environ["RENPY_MULTIPERSISTENT"]) ]
-
     # Make the new persistent directory, why not?
     try:
         os.makedirs(files[-1]) # type: ignore
     except Exception:
         pass
 
-    fn = "" # prevent a warning from happening.
     data = None
 
     # Find the first file that actually exists. Otherwise, use the last
@@ -516,7 +588,8 @@ def MultiPersistent(name, save_on_quit=False):
         fn = os.path.join(fn, name) # type: ignore
         if os.path.isfile(fn):
             try:
-                data = open(fn, "rb").read()
+                with open(fn, "rb") as mpf:
+                    data = mpf.read()
                 break
             except Exception:
                 pass
@@ -527,13 +600,14 @@ def MultiPersistent(name, save_on_quit=False):
         try:
             rv = loads(data)
         except Exception:
-            renpy.display.log.write("Loading MultiPersistent at %r:" % fn)
+            renpy.display.log.write("Loading MultiPersistent at %r:" % fn) # type: ignore
             renpy.display.log.exception()
 
-    rv._filename = fn
+    rv._filename = fn # type: ignore
+    rv._name = name
+    rv._save_on_quit = save_on_quit
 
-    if save_on_quit:
-        save_MP_instances.add(rv)
+    MP_instances.add(rv)
 
     return rv
 
