@@ -1,4 +1,4 @@
-# Copyright 2004-2022 Tom Rothamel <pytom@bishoujo.us>
+# Copyright 2004-2024 Tom Rothamel <pytom@bishoujo.us>
 #
 # Permission is hereby granted, free of charge, to any person
 # obtaining a copy of this software and associated documentation files
@@ -88,7 +88,7 @@ def mutator(method):
 
         global mutate_flag
 
-        mutated = renpy.game.log.mutated # @UndefinedVariable
+        mutated = renpy.game.log.mutated
 
         if id(self) not in mutated:
             mutated[id(self)] = (weakref.ref(self), self._clean())
@@ -193,7 +193,7 @@ class RevertableList(list):
     append = mutator(list.append)
     extend = mutator(list.extend)
     insert = mutator(list.insert)
-    pop = mutator(list.pop) # type: ignore
+    pop = mutator(list.pop)
     remove = mutator(list.remove)
     reverse = mutator(list.reverse)
     sort = mutator(list.sort)
@@ -203,6 +203,8 @@ class RevertableList(list):
         @_method_wrapper(method)
         def newmethod(*args, **kwargs):
             l = method(*args, **kwargs) # type: ignore
+            if l is NotImplemented:
+                return l
             return RevertableList(l)
 
         return newmethod
@@ -210,6 +212,8 @@ class RevertableList(list):
     __add__ = wrapper(list.__add__) # type: ignore
     if PY2:
         __getslice__ = wrapper(list.__getslice__) # type: ignore
+    __mul__ = wrapper(list.__mul__) # type: ignore
+    __rmul__ = wrapper(list.__rmul__) # type: ignore
 
     del wrapper
 
@@ -221,19 +225,11 @@ class RevertableList(list):
         else:
             return rv
 
-    def __mul__(self, other):
-        if not isinstance(other, int):
-            raise TypeError("can't multiply sequence by non-int of type '{}'.".format(type(other).__name__))
-
-        return RevertableList(list.__mul__(self, other))
-
-    __rmul__ = __mul__ # type: ignore
-
     def copy(self):
         return self[:]
 
     def clear(self):
-        self[:] = []
+        del self[:]
 
     def _clean(self):
         """
@@ -333,6 +329,25 @@ class RevertableDict(dict):
         def has_key(self, key):
             return (key in self)
 
+        # https://peps.python.org/pep-0584 methods
+        def __or__(self, other):
+            if not isinstance(other, dict):
+                return NotImplemented
+            rv = RevertableDict(self)
+            rv.update(other)
+            return rv
+
+        def __ror__(self, other):
+            if not isinstance(other, dict):
+                return NotImplemented
+            rv = RevertableDict(other)
+            rv.update(self)
+            return rv
+
+        def __ior__(self, other):
+            self.update(other)
+            return self
+
     def copy(self):
         rv = RevertableDict()
         rv.update(self)
@@ -351,6 +366,35 @@ class RevertableDict(dict):
             self[k] = v
 
 
+class RevertableDefaultDict(RevertableDict):
+    """
+    :doc: rollbackclasses
+    :name: defaultdict
+    :args: (default_factory, /, *args, **kwargs)
+
+    This is a revertable version of collections.defaultdict. It takes a
+    factory function. If a key is accessed that does not exist, the `default_factory`
+    function is called with the key as an argument, and the result is
+    returned.
+
+    While the default_factory attribute is present on this object, it does not
+    participate in rollback, and so should not be changed.
+    """
+
+    def __init__(self, default_factory=None, *args, **kwargs):
+        self.default_factory = default_factory
+        super(RevertableDefaultDict, self).__init__(*args, **kwargs)
+
+    def __missing__(self, key):
+        if self.default_factory is None:
+            raise KeyError(key)
+
+        rv = self.default_factory()
+        self[key] = rv
+        return rv
+
+
+
 class RevertableSet(set):
 
     def __setstate__(self, state):
@@ -448,11 +492,11 @@ class RevertableObject(object):
 
     def __init_subclass__(cls):
         if renpy.config.developer and "__slots__" in cls.__dict__:
-            raise TypeError("Classes with __slots__ do not support rollback."
+            raise TypeError("Classes with __slots__ do not support rollback. "
                             "To create a class with slots, inherit from python_object instead.")
 
-    __setattr__ = mutator(object.__setattr__) # type: ignore
-    __delattr__ = mutator(object.__delattr__) # type: ignore
+    __setattr__ = mutator(object.__setattr__)
+    __delattr__ = mutator(object.__delattr__)
 
     def _clean(self):
         return self.__dict__.copy()
@@ -465,6 +509,50 @@ class RevertableObject(object):
         self.__dict__.update(compressed)
 
 
+class MultiRevertable(object):
+    """
+    :doc: rollbackclasses
+    :name: MultiRevertable
+
+    MultiRevertable is a mixin class that allows an object to inherit
+    from more than one kind of revertable object. To use it, from MultiRevertable,
+    then from the revertable classes you want to inherit from.
+
+    For example::
+
+        class MyDict(MultiRevertable, dict, object):
+            pass
+
+    will create an class that will rollback both its dict contents and
+    object fields.
+    """
+
+    def _rollback_types(self):
+        rv = [ ]
+
+        for i in self.__class__.__mro__:
+
+            if i is MultiRevertable:
+                continue
+
+            if "_rollback" in i.__dict__:
+                rv.append(i)
+
+        return rv
+
+    def _clean(self):
+        return tuple(i._clean(self) for i in self._rollback_types())
+
+    def _compress(self, clean):
+        return tuple(i._compress(self, c) for i, c in zip(self._rollback_types(), clean))
+
+    def _rollback(self, compressed):
+
+        for i, c in zip(self._rollback_types(), compressed):
+            i._rollback(self, c)
+
+
+
 def checkpointing(method):
 
     @_method_wrapper(method)
@@ -477,6 +565,16 @@ def checkpointing(method):
     return do_checkpoint
 
 
+def list_wrapper(method):
+
+    @_method_wrapper(method)
+    def newmethod(*args, **kwargs):
+        l = method(*args, **kwargs)
+        return RevertableList(l)
+
+    return newmethod
+
+
 class RollbackRandom(random.Random):
     """
     This is used for Random objects returned by renpy.random.Random.
@@ -503,9 +601,12 @@ class RollbackRandom(random.Random):
 
     if PY2:
         jumpahead = checkpointing(mutator(random.Random.jumpahead)) # type: ignore
+    else:
+        choices = list_wrapper(random.Random.choices)
+    sample = list_wrapper(random.Random.sample)
 
     getrandbits = checkpointing(mutator(random.Random.getrandbits))
-    seed = checkpointing(mutator(random.Random.seed)) # type: ignore
+    seed = checkpointing(mutator(random.Random.seed))
     random = checkpointing(mutator(random.Random.random))
 
     def Random(self, seed=None):
@@ -529,6 +630,10 @@ class DetRandom(random.Random):
         super(DetRandom, self).__init__()
         self.stack = [ ]
 
+    if not PY2:
+        choices = list_wrapper(random.Random.choices)
+    sample = list_wrapper(random.Random.sample)
+
     def random(self):
 
         if self.stack:
