@@ -56,7 +56,6 @@ def parse_var_node(env, sig, signode):
 
 style_seen_ids = set()
 
-
 def parse_style_node(env, sig, signode):
     m = re.match(r'(\S+)(.*)', sig)
 
@@ -78,6 +77,25 @@ def parse_style_node(env, sig, signode):
     return ref
 
 
+scpref_seen_ids = set()
+
+def parse_scpref_node(env, sig, signode):
+    m = re.match(r'(\S+)(.*)', sig)
+
+    signode += sphinx.addnodes.desc_name(m.group(1), m.group(1))
+    signode += docutils.nodes.Text(m.group(2), m.group(2))
+
+    ref = m.group(1)
+
+    while ref in scpref_seen_ids:
+        print("duplicate id:", ref)
+        ref = ref + "_alt"
+
+    scpref_seen_ids.add(ref)
+
+    return ref
+
+
 class PythonIndex(sphinx.domains.Index):
     name = "function-class-index"
     localname = "Function and Class Index"
@@ -183,6 +201,7 @@ def setup(app):
     app.add_object_type("var", "var", "single: %s (variable)", parse_node=parse_var_node)
     app.add_object_type("style-property", "propref", "single: %s (style property)", parse_node=parse_style_node)
     app.add_object_type("transform-property", "tpref", "single: %s (transform property)")
+    app.add_object_type("screen-property", "scpref", "single: %s (screen property)", parse_node=parse_scpref_node)
     app.add_object_type("text-tag", "tt", "single: %s (text tag)")
 
     add_index(app, "std", "style-property", "Style Property Index")
